import 'dart:convert';
import 'package:bento_support_flutter/controllers/loader.controller.dart';
import 'package:bento_support_flutter/models/client.model.dart';
import 'package:bento_support_flutter/models/organization.model.dart';
import 'package:bento_support_flutter/services/message.service.dart';
import 'package:bento_support_flutter/services/parser.service.dart';
import 'package:collection/collection.dart';
import 'package:bento_support_flutter/controllers/auth_controller.dart';
import 'package:bento_support_flutter/controllers/socket.controller.dart';
import 'package:bento_support_flutter/models/complaint.dart';
import 'package:bento_support_flutter/models/dtos/create_complaint_dto.dart';
import 'package:bento_support_flutter/models/pagination.dart';
import 'package:bento_support_flutter/models/socket_message.dart';
import 'package:bento_support_flutter/services/complaint.service.dart';
import 'package:bento_support_flutter/services/socket.service.dart';
import 'package:file_picker/file_picker.dart';
import 'package:flutter/material.dart';
import 'package:get/get_instance/src/extension_instance.dart';
import 'package:get/get_state_manager/get_state_manager.dart';
import 'package:get/state_manager.dart';
import 'package:pull_to_refresh/pull_to_refresh.dart';
import 'package:rothko_mobile/rothko_system/snackbar/snackbar_route.dart';

class ChatController extends GetxController {
  ScrollController chatScreenScrollController = ScrollController();

  late BuildContext context;

  FileType pickingType = FileType.any;

  RxBool expanded = RxBool(false);

  Rx<ChatState> currentState = ChatState.initial.obs;

  RxBool isResolved = RxBool(false);
  RxBool showEmoji = RxBool(false);

  late Rx<Profile> currentAgent;

  late Rx<Complaint> lastComplaint;

  late Function _onLogin;

  RxList<Rx<Message<Profile>>> messages = <Rx<Message<Profile>>>[].obs;

  Rx<PaginationMetadata> messagePaginationMetadata =
      PaginationMetadata(page: 0, pageCount: 0, perPage: 20, total: 0).obs;

  var conversationViewController = RefreshController();

  void initializeLoginUserView() {
    LoaderController.instance.loadAsyncProcess(() async {
      var authController = Get.find<AuthController>();
      Get.find<SocketController>().initialize();
      var complaintResp = await ComplaintService.fetchClientLastComplaint(
          authController.client.value.id);
      if (complaintResp.data != null) {
        lastComplaint = (complaintResp.data as Complaint).obs;
        fetchMessages();
        currentState.value = ChatState.startConversationView;
      } else {
        // Move to the select tag view
        currentState.value = ChatState.selectTagView;
      }
      fetchMessages();
    });
  }

  void initializeAnonymousUserView(Responder responder) async {
    // set the first default message
    addMessage(Message<Profile>.fromMap(<String, dynamic>{
      'deliveredTo': <dynamic>[],
      'readBy': <dynamic>[],
      'attachments': <dynamic>[],
      'type': 'new.message',
      'sender': 'user',
      'complaint': 'anonymous',
      'sentBy': <String, dynamic>{
        'profileImageUrl': responder.avatarUrl,
        'name': responder.name,
        'organization': '6155b76eecc52777f849ef1f',
        'clientType': 'user',
        'id': ''
      },
      'body': responder.welcomeMessage,
      'createdAt': DateTime.now().toIso8601String(),
      'updatedAt': DateTime.now().toIso8601String(),
      'id': '',
      'uuid': '',
    }));
    currentState.value = ChatState.anonymousMainView;
  }

  void fetchMessages({bool loadPreviousMessages = false}) async {
    if (!loadPreviousMessages) {
      messagePaginationMetadata.value =
          messagePaginationMetadata.value.copyWith(page: 0);
    }

    var resp = await MessageService.fetchMessages(
        Get.find<AuthController>().client.value.id,
        messagePaginationMetadata.value);
    if (resp.meta.pageCount != 0) {
      var data = resp.data.map((e) => e.obs);
      if (!loadPreviousMessages) {
        messages.clear();
      }
      messages.addAll(data);
      messages.sort((a, b) => a.value.createdAt.compareTo(b.value.createdAt));
      messagePaginationMetadata.value = resp.meta as PaginationMetadata;
    }
    if (loadPreviousMessages) {
      conversationViewController.refreshCompleted();
    } else {
      conversationViewController.loadComplete();
      moveToLastMessage();
    }
  }

  void startConversationView({Client? client, List<String>? tags}) {
    LoaderController.instance.loadAsyncProcess(() async {
      var complaint = await createComplaint(client: client, tags: tags);

      Get.find<AuthController>().client =
          (complaint.client as Client<String>).obs;
      lastComplaint = complaint.obs;

      Get.find<SocketController>().initialize();
      currentState.value = ChatState.startConversationView;
    });
  }

  Future<Complaint> createComplaint(
      {Client? client, List<String>? tags}) async {
    var resp = await ComplaintService.createComplaint(CreateComplaintDto(
        name: client?.name ?? 'New User',
        email: client?.email ?? '',
        phoneNumber: client?.phoneNumber ?? '',
        clientReference: client?.clientReference,
        tags: tags ?? []));
    return resp.data;
  }

  void setLoginAction(Function action) {
    _onLogin = action;
  }

  void login() {
    expanded.value = false;
    _onLogin();
  }

  void toggleEmoji() {
    showEmoji.value = !showEmoji.value;
  }

  void openFileExplorer() async {
    var selectedFiles = (await FilePicker.platform.pickFiles(
          type: FileType.any,
          allowMultiple: true,
        ))
            ?.files ??
        [];
    if (selectedFiles.isEmpty) {
      return;
    }
    sendNewMessage(
        body: '',
        attachmentPaths: selectedFiles
            .map((e) => MessageAttachment(filename: e.name, file: e))
            .toList());
  }

  void handleOnMessage(SocketMessage data) {
    switch (data.event) {
      case RoomEventType.newComplaint:
        {
          handleOnNewMessage(data);
          break;
        }
      case RoomEventType.newMessage:
        {
          handleOnNewMessage(data);
          break;
        }
      case RoomEventType.messageRead:
        {
          break;
        }
      case RoomEventType.messageDelivered:
        {
          break;
        }
      case RoomEventType.leftRoom:
        {
          break;
        }
      case RoomEventType.complaintRated:
        {
          break;
        }
      case RoomEventType.complaintResolved:
        {
          handleOnNewMessage(data);
          break;
        }
    }
  }

  void handleOnNewMessage(SocketMessage data) {
    var message =
        data.message.copyWith<Profile>(sentBy: data.message.sentBy as Profile);
    // check if message already exists
    var messageExist = messages.reversed
        .firstWhereOrNull((element) => element.value.id == message.id);
    if (messageExist == null) {
      messages.add(message.obs);
    } else {
      var index = messages.indexOf(messageExist);
      messages[index].value = data.message
          .copyWith<Profile>(sentBy: data.message.sentBy as Profile);
    }
  }

  Rx<Message<dynamic>> addMessage(Message<Profile> message) {
    var reactiveMessage = message.obs;
    messages.add(reactiveMessage);
    moveToLastMessage();
    return reactiveMessage;
  }

  void moveToLastMessage() {
    if (chatScreenScrollController.hasClients) {
      Future.delayed(Duration.zero, () {
        chatScreenScrollController
            .jumpTo(chatScreenScrollController.position.maxScrollExtent + 100);
      });
    }
  }

  void reviewMessage(
      Message<Profile> message, int review, BuildContext context) {
    if (message.review != null) {
      ScaffoldMessenger.of(context).showSnackBar(
          SnackBar(content: Text('You have already rated this complaint')));
      return;
    }
    var authController = Get.find<AuthController>();
    var updatedMessage = message.copyWith<String>(
      review: review,
      sentBy: authController.client.value.id,
      type: MessageType.complaintRated,
    );
    var socketMessage = SocketMessage(
        room: getClientRoom,
        event: RoomEventType.newMessage,
        message: updatedMessage);

    Get.find<SocketController>().emitEvent(ON_MESSAGE, socketMessage);

    var handler =
        messages.firstWhereOrNull((element) => element.value.id == message.id);
    if (handler != null) {
      handler.value.review = review;
    }
  }

  void sendNewMessage(
      {required String body,
      List<MessageAttachment> attachmentPaths = const <MessageAttachment>[]}) {
    LoaderController.instance.loadAsyncProcess(() {
      var uuid = UniqueKey().toString();
      var authController = Get.find<AuthController>();
      var socketMessage = SocketMessage(
          room: getClientRoom,
          event: RoomEventType.newMessage,
          message: Message<String>(
            deliveredTo: <dynamic>[],
            readBy: <dynamic>[],
            attachments: attachmentPaths,
            complaint: lastComplaint.value.id,
            type: MessageType.newMessage,
            sender: MessageParticipantType.client,
            sentBy: authController.client.value.id,
            createdAt: DateTime.now(),
            updatedAt: DateTime.now(),
            body: body,
            uuid: uuid,
          ));

      var messageWithSentByPopulated = socketMessage.message.copyWith<Profile>(
          sentBy: Profile(
              profileImageUrl: authController.client.value.profileImageUrl,
              name: authController.client.value.name,
              id: authController.client.value.id));

      var messageHandle = addMessage(messageWithSentByPopulated);

      if (socketMessage.message.attachments.isNotEmpty) {
        Future.wait(socketMessage.message.attachments.map((e) async {
          Future.delayed(Duration(seconds: 4), () {});
          var data = await ParserService.getBytesFromFile(e.file!);
          var urlResp = await MessageService.uploadFile(
            base64.encode(data),
            e.filename,
            e.file!.extension!,
            e.mimeType,
          );

          e.url = urlResp.data;
          e.loadingState = LoadingState.success;
          return e;
        })).then((value) {
          messageWithSentByPopulated.attachments = value;
          messageHandle.value = messageWithSentByPopulated;
          Get.find<SocketController>().emitEvent(ON_MESSAGE, socketMessage);
        });
      } else {
        Get.find<SocketController>().emitEvent(ON_MESSAGE, socketMessage);
      }
    }, blocking: false);
  }

  String get getClientRoom {
    var authController = Get.find<AuthController>();
    return 'client-' +
        (authController.client.value.organization) +
        '-' +
        authController.client.value.id;
  }
}

enum ChatState {
  anonymousMainView,
  startConversationView,
  selectTagView,
  initial
}
