import 'dart:async';
import 'dart:convert';

import 'package:bento_support_flutter/services/api/api_request_type.dart';
import 'package:bento_support_flutter/services/api/api_response.dart';
import 'package:bento_support_flutter/services/config.dart';
import 'package:get/get.dart';
import 'package:http/http.dart' as http;

class ApiServices {
  /// A constant list of status codes that indicate an http reuest is successesfull.
  static const successStatusCodes = [201, 200];

  static late String publicKey;

  static late String clientReference;

  /// Get the Lendly api-header.
  static Map<String, String> get support_api_header {
    return <String, String>{
      'public_key': publicKey,
      'Content-Type': 'application/json',
    };
  }

  /// Make an http `GET` request (call) to the api.
  static Future<ApiResponse<T>> get<T>({
    required String path,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    return call<T>(
      ApiRequestType.get,
      path: path,
      queryParameters: queryParameters,
      toModel: toModel,
    );
  }

  /// Make an http `POST` request (call) to the api.
  static Future<ApiResponse<T>> post<T>({
    required String path,
    required Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    return call<T>(
      ApiRequestType.post,
      path: path,
      data: data,
      queryParameters: queryParameters,
      toModel: toModel,
    );
  }

  /// Make an http `PUT` request (call) to the api.
  static Future<ApiResponse<T>> put<T extends Object>({
    required String path,
    required Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    return call<T>(
      ApiRequestType.put,
      path: path,
      data: data,
      queryParameters: queryParameters,
      toModel: toModel,
    );
  }

  /// Make an http `PATCH` request (call) to the api.
  static Future<ApiResponse<T>> patch<T extends Object>({
    required String path,
    required Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    return call<T>(
      ApiRequestType.patch,
      path: path,
      data: data,
      queryParameters: queryParameters,
      toModel: toModel,
    );
  }

  /// Make an http `DELETE` request (call) to the api.
  static Future<ApiResponse<T>> delete<T>({
    required String path,
    required Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    return call<T>(
      ApiRequestType.delete,
      path: path,
      data: data,
      queryParameters: queryParameters,
      toModel: toModel,
    );
  }

  /// Make an http request (call) to the api.
  ///
  /// Check out the following method which uses [call] underneath to make various distinguished http request.
  ///   * [get] to make a get request,
  ///   * [post] to make a post request,
  ///   * [put] to make a put request,
  ///   * [patch] to make a patch request,
  ///   * [delete] to make a delete request.
  static Future<ApiResponse<T>> call<T>(
    ApiRequestType requestType, {
    required String path,
    Map<String, dynamic>? data,
    Map<String, dynamic>? queryParameters,
    required T Function(dynamic object) toModel,
    String? baseUrl,
  }) async {
    await Future<dynamic>.delayed(
      Duration(seconds: 3),
    );
    late final http.Response resp;

    var iBaseUrl = baseUrl ?? EnvConfig.baseApiUrl;

    switch (requestType) {
      case ApiRequestType.get:
        resp = await http.get(
          Uri.parse(iBaseUrl + path),
          headers: support_api_header,
        );
        break;
      case ApiRequestType.post:
        final _body = data != null ? jsonEncode(data) : data;
        resp = await http.post(
          Uri.parse(iBaseUrl + path),
          headers: support_api_header,
          body: _body,
        );
        break;
      case ApiRequestType.put:
        final _body = data != null ? jsonEncode(data) : data;
        resp = await http.put(
          Uri.parse(iBaseUrl + path),
          headers: support_api_header,
          body: _body,
        );
        break;
      case ApiRequestType.patch:
        final _body = data != null ? jsonEncode(data) : data;
        resp = await http.patch(
          Uri.parse(iBaseUrl + path),
          headers: support_api_header,
          body: _body,
        );
        break;
      case ApiRequestType.delete:
        final _body = data != null ? jsonEncode(data) : data;
        resp = await http.delete(
          Uri.parse(iBaseUrl + path),
          headers: support_api_header,
          body: _body,
        );
        break;
      default:
        throw Exception('Unknown request type');
    }

    if (resp.statusCode == 401) {
      throw Exception('Unauthorized');
    }

    final decodedResp = jsonDecode(resp.body) as Map<String, dynamic>;
    var _apiRes = ApiResponse.fromJson<dynamic>(decodedResp);

    if (!successStatusCodes.contains(resp.statusCode)) {
      throw Exception(_apiRes.message);
    }

    final dynamic _oldResData = _apiRes.data;
    late final _finalRes = _apiRes.copyWith<T>(
      success: true,
      data: toModel(_oldResData),
    );

    return _finalRes;
  }
}
